"""Control flow graphs

- Label and its subclasses describe the label on the edges of the CfGs
- Nodes describes nodes, with an id and its successors
- Graph describes a graph
- Cfg describes a cfg, i.e. a graph with initial, final, and error nodes
- function cfg creates a cfg for the given program
"""
from syntax import *
from dataclasses import dataclass

@dataclass
class Label:
    """generic class for edges' labels"""
    pass

@dataclass
class LSkip(Label):
    """label skip (does nothing)"""
    is_back_edge: bool
    def __str__(self):
        return "skip"

@dataclass
class LSet(Label):
    """assignment label"""
    var: str
    expr: ArithExpr
    def __str__(self):
        return f"{self.var} := {self.expr}"

@dataclass
class LTest(Label):
    """test label"""
    cond: BoolExpr
    def __str__(self):
        return f"?{self.cond}"

@dataclass
class LErr(Label):
    """error label"""
    err: Expr
    def __str__(self):
        return f"Err({self.err})"

class Node:
    """node of the graph, with its id and its successors"""
    def __init__(self,nid):
        self.nid = nid
        self.succs = []
    def __str__(self):
        return f"Node {self.nid}"

class Graph:
    nb_id: int
    nodes: list[Node]

    """a graph, i.e. a list of nodes"""
    def __init__(self):
        self.nb_nid = 0
        self.nodes = []

    def new_node(self) -> Node:
        """creates a new node, with a fresh id and no succs"""
        node = Node(self.nb_nid)
        self.nb_nid += 1
        self.nodes += [node]
        return node

    def new_edge(self,n1: Node,n2: Node,l: Label) -> None:
        """creates a new edge.
        If the nodes are not already in the graph, they are added.
        In that case, no check is made that a node with the same id
        is already present or not.
        """
        if n1 not in self.nodes:
            self.nodes += [n1]
        if n2 not in self.nodes:
            self.nodes += [n2]
        n1.succs += [(n2,l)]

def cfg_aux(g: "Cfg", e: Node, i: Node, instr: Instr) -> Node:
    match instr:
        case ISkip():
            f = g.new_node()
            g.new_edge(i, f, LSkip(is_back_edge=False))
            return f

        case ISet(v, expr):
            f = g.new_node()
            g.new_edge(i, f, LSet(v, expr))
            g.new_edge(i, e, LErr(expr))
            return f

        case ISeq(i1, i2):
            m = cfg_aux(g, e, i, i1)
            return cfg_aux(g, e, m, i2)

        case ICond(cond, i1, i2):
            f = g.new_node()
            g.new_edge(i, f, LTest(BENeg(cond)))
            g.new_edge(i, e, LErr(cond))
            t = cfg_aux(g, e, f, i1)
            g.new_edge(i, f, LTest(cond))
            return cfg_aux(g, e, t, i2)

        case ILoop(cond, body):
            ib = g.new_node()
            f = g.new_node()
            g.new_edge(i, ib, LTest(cond))          # vers boucle si cond vrai
            g.new_edge(i, f, LTest(BENeg(cond)))    # sortie si cond faux
            g.new_edge(i, e, LErr(cond))            # erreur si test échoue
            fb = cfg_aux(g, e, ib, body)
            g.new_edge(fb, i, LSkip(is_back_edge=True))  # <-- arc de retour
            return f


class Cfg:
    """describes a CfG"""
    g: Graph
    err_node: Node
    init_node: Node
    final_node: Node
    def __init__(self,instr):
        self.g = Graph()
        self.err_node = self.g.new_node()
        self.init_node = self.g.new_node()
        self.final_node = cfg_aux(self.g,self.err_node,self.init_node,instr)

    def __str__(self):
        res = ""
        for node in self.g.nodes:
            res+=f"{node}:\n"
            for (dest,lab) in node.succs:
                res+=f"  {lab} --> {dest}\n"
        res+=f"init node is {self.init_node}\n"
        res+=f"final node is {self.final_node}\n"
        res+=f"err node is {self.err_node}\n"
        return res
